import json

try:
    import bpy
    import mathutils
    from . import scene
    from .. import utils
    from .. import config
    from ..tools import bouncecurve as bc
    from ..bouncecurve import bcutils
except ModuleNotFoundError as e:
    import scene
    import utils
    import config
    import bcutils
    from tools import bouncecurve as bc


def generate_bounce_curve(context, points, factor):
    return bc.create(context, points, factor)



class GenerateBounceCurveOperator(bpy.types.Operator):
    bl_idname = "object.bc_gen_bounce_curve"
    bl_label = "生成弹跳曲线/动画"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        data = context.scene.bc_score_data
        try:
            j = json.loads(data)
        except:
            self.report({'ERROR'}, '乐谱数据丢失，请重新生成乐谱')
            return {'FINISHED'}
        obj_score = context.scene.bc_obj_svg_score
        if not obj_score:
            self.report({'INFO'}, '请选择乐谱')
            return {'FINISHED'}
        obj_bullet = context.scene.bc_obj_bounce
        if not obj_bullet:
            self.report({'INFO'}, '选择弹跳对象才会自动生成动画')
        parts = j.get('parts', [])
        context.scene.bc_real_obj_bounce.clear()
        col_bullet = utils.get_or_new_collection('子弹')
        col_curve = utils.get_or_new_collection('跳线')
        for part in parts:
            times = [t.get('onset_time') for t in part.get('times', [])]
            note_paths = part.get('note_paths', [])
            if len(times) < 1:
                self.report({'INFO'}, '乐谱数据异常1')
                continue
            for path in note_paths:
                if len(times) != len(path):
                    print('times len: {}, chords len:{}'.format(len(times), len(path)))
                    self.report({'INFO'}, '乐谱数据异常2')
                    continue
                point_indices = []
                tmp_points = []
                for point_ind, point in enumerate(path):
                    if point is None:
                        continue
                    point_indices.append(point_ind)
                    tmp_points.append(mathutils.Vector(point))

                points = [mathutils.Vector((p.x, p.y, 0)) for p in tmp_points]
                points.reverse()
                curve = generate_bounce_curve(self, points, context.scene.bc_bounce_factor)
                utils.move_object(col_curve, curve)
                if not obj_bullet:
                    continue

                obj_bullet_copy = utils.copy_obj(obj_bullet)
                utils.move_object(col_bullet, obj_bullet_copy)
                item = context.scene.bc_real_obj_bounce.add()
                item.obj = obj_bullet_copy
                modifier = obj_bullet_copy.modifiers.new('curve', 'CURVE')
                modifier.object = curve
                modifier.deform_axis = 'POS_X'

                obj_bullet_copy.location = (0, 0, 0)
                obj_bullet_copy.lock_location[1] = True
                obj_bullet_copy.lock_location[2] = True

                floor_dists = bcutils.calculate_curve_section_length(curve)
                for i, dist in enumerate(floor_dists):
                    frame = int(times[point_indices[i]] * bpy.context.scene.render.fps)
                    # print('frame: {} distance:{}'.format(frame, dist))
                    utils.set_keyframe(obj_bullet_copy, 'location', frame, (dist, 0, 0))
                for curve in obj_bullet_copy.animation_data.action.fcurves:
                    if curve.data_path == 'location':
                        for kf in curve.keyframe_points:
                            kf.interpolation = 'LINEAR'
        return {'FINISHED'}
