import base64
import json
import math
import os
import tempfile

try:
    import bpy
    import mathutils
    from . import scene
    from .. import utils
    from .. import config
    from ..tools import bouncecurve as bc
    from ..bouncecurve import bcutils
except ModuleNotFoundError as e:
    import scene
    import utils
    import config
    import bcutils
    from tools import bouncecurve as bc


def get_strokes():
    obj = bpy.context.active_object
    if not obj:
        return []
    if bpy.app.version >= (4, 3, 0):
        if obj.type != 'GREASEPENCIL':
            return []
    else:
        if obj.type != 'GPENCIL':
            return []
    if bpy.app.version < (4, 2, 0):
        bpy.ops.object.mode_set(mode='EDIT_GPENCIL')
    elif bpy.app.version >= (4, 3, 0):
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.editmode_toggle()
    else:
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.gpencil.editmode_toggle()
    if len(obj.data.layers) < 1:
        return []
    layer = obj.data.layers[0]
    if len(layer.frames) < 1:
        return []
    f = layer.frames[0]

    strokes = []
    if bpy.app.version >= (4, 3, 0):
        for s in f.drawing.strokes:
            points = []
            for p in s.points:
                points.append((p.position[0], p.position[1], p.position[2]))
            strokes.append(points)
    else:
        for s in f.strokes:
            points = []
            for p in s.points:
                points.append((p.co[0], p.co[1], p.co[2]))
            strokes.append(points)
    bpy.ops.object.mode_set(mode='OBJECT')
    return strokes


def get_score_staff_lines(col_name):
    objs = bpy.data.collections[col_name].objects[:5]
    return [
        (obj.data.splines[0].bezier_points[0].co[:2], obj.data.splines[0].bezier_points[1].co[:2]) for obj in objs
    ]


def select_all_objects(col):
    for child_col in col.children:
        select_all_objects(child_col)
    for obj in col.objects:
        obj.select_set(True)


def group_objects(context, data, col):
    scale = data['scale']
    offset = data['offset']
    obj_scales = context.scene.bc_obj_svg_score.scale
    col_others = utils.get_or_new_collection('其他', col)
    col_others_col = utils.get_or_new_collection('竖线', col_others)
    # group notes
    for part_ind, part in enumerate(data['parts']):
        col_part = utils.get_or_new_collection(f'音符{part_ind + 1}', col)
        note_paths = part['note_paths']
        note_paths.append([loc for locs in part['tie_locations'] for loc in locs])
        for note_path in note_paths:
            note_path = [point for point in note_path]
            for obj in col.objects:
                if len(obj.data.splines[0].bezier_points) != 4:
                    continue
                for ind_real_point, real_point in enumerate(note_path):
                    if abs(obj.location[0] - real_point[0]) > 0.03 or abs(obj.location[1] - real_point[1]) > 0.015:
                        continue
                    utils.move_object(col_part, obj)
                    note_path.pop(ind_real_point)
                    break
    # group staff lines
    col_staff = utils.get_or_new_collection('谱线', col)
    col_staff_row = utils.get_or_new_collection('横线', col_staff)
    col_staff_col = utils.get_or_new_collection('竖线', col_staff)
    staff_ys = set()
    for obj in col.objects:
        points = obj.data.splines[0].bezier_points
        if len(points) != 2:
            continue
        if abs(points[0].co[0]) < 0.001:
            continue
        if abs(points[0].co[1]) < 0.002 and abs(points[1].co[1]) < 0.002:
            staff_ys.add(obj.location[1])
            utils.move_object(col_staff_row, obj)
    for obj in col.objects:
        points = obj.data.splines[0].bezier_points
        if len(points) != 2:
            continue
        if abs(points[0].co[1]) < 0.005:
            continue
        loc = obj.location
        match_count = 0
        for point in points:
            for staff_y in staff_ys:
                if abs(loc[1] + point.co[1] * obj.scale[1] - staff_y) < 0.02:
                    match_count += 1
                    break
            if match_count == 2:
                utils.move_object(col_staff_col, obj)
                break
        if match_count < 2:
            utils.move_object(col_others_col, obj)

    # group others
    for obj in col.objects:
        utils.move_object(col_others, obj)


class GenerateScoreOperator(bpy.types.Operator):
    bl_idname = "object.bc_gen_score"
    bl_label = "生成乐谱"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        # check if it's midi file
        if (len(context.scene.bc_midi) < 1
                or not context.scene.bc_midi.endswith('.mid')
        ):
            self.report({'ERROR'}, '请选择mid文件')
            return {'FINISHED'}
        if not os.path.exists(context.scene.bc_midi) or not os.path.isfile(context.scene.bc_midi):
            print(context.scene.bc_midi)
            self.report({'ERROR'}, 'mid文件不存在')
            return {'FINISHED'}
        key = context.scene.mc_auth_key.strip()
        secret = context.scene.mc_auth_secret.strip()
        result, msg, data, meta = utils.analyze_mid(key, secret, context.scene.bc_midi)
        if not result:
            self.report({'ERROR'}, '生成失败' if msg is None else msg)
            return {'FINISHED'}
        staff_lines = data.get('staff_lines', [])

        if len(staff_lines) < 2:
            self.report({'INFO'}, '无法确定乐谱坐标')
            return {'FINISHED'}
        parts = data.get('parts', [])
        for p in parts:
            times = p.get('times', [])
            note_paths = p.get('note_paths', [])
            if len(times) < 1:
                self.report({'INFO'}, '乐谱数据异常1')
                continue
        svg = data.get('svg')

        with tempfile.TemporaryDirectory() as tmpdir:
            name = context.scene.bc_midi.replace('.mid', '.svg')
            filename = os.path.basename(name)
            dest_file = os.path.join(tmpdir, filename)
            with open(dest_file, 'wb') as f:
                f.write(base64.b64decode(svg))

            if context.scene.bc_import_svg_as_gspencil:
                if 'gpencil_import_svg' in dir(bpy.ops.wm):
                    try:
                        bpy.ops.wm.gpencil_import_svg(filepath="", directory=os.path.dirname(dest_file),
                                                      files=[{"name": os.path.basename(dest_file)}])
                    except:
                        pass
                elif 'grease_pencil_import_svg' in dir(bpy.ops.wm):
                    try:
                        bpy.ops.wm.grease_pencil_import_svg(filepath="", directory=os.path.dirname(dest_file),
                                                            files=[{"name": os.path.basename(dest_file)}])
                    except:
                        pass
                else:
                    self.report({'ERROR'}, '该 Blender 版本无法生成蜡笔乐谱，请使用合适版本')
                    return {'FINISHED'}
                context.scene.bc_obj_svg_score = bpy.context.active_object
                context.scene.bc_obj_svg_score.rotation_euler = (math.radians(270), 0, 0)
                context.scene.bc_obj_svg_score.scale = (20, 20, 20)
                scale, offset = bcutils.cal_scale_and_offset(staff_lines, get_strokes()[:5])
            else:
                bpy.ops.import_curve.svg(filepath=dest_file)
                scale = 40
                for obj in bpy.data.collections[filename].objects:
                    obj.scale = (scale, scale, scale)
                col = bpy.data.collections[filename]
                context.scene.bc_obj_svg_score = col.objects[0]
                scale, offset = bcutils.cal_scale_and_offset(staff_lines, get_score_staff_lines(filename))
        del data['svg']
        data['scale'] = scale
        data['offset'] = offset
        obj_scales = context.scene.bc_obj_svg_score.scale
        for part in data.get('parts', []):
            real_note_paths = [[list((mathutils.Vector(point) * mathutils.Vector(scale) + mathutils.Vector(
                offset)) * mathutils.Vector((obj_scales[0], obj_scales[2]))) for point in note_path] for note_path in
                               part.get('note_paths', [[]])]
            part['note_paths'] = real_note_paths
            real_tie_locations = [[list((mathutils.Vector(point) * mathutils.Vector(scale) + mathutils.Vector(
                offset)) * mathutils.Vector((obj_scales[0], obj_scales[2]))) for point in note_path] for note_path in
                                  part.get('tie_locations', [[]])]
            part['tie_locations'] = real_tie_locations

        context.scene.bc_score_data = json.dumps(data)
        context.scene.mc_expire_tips = meta.get('expired_at', '')

        if not context.scene.bc_import_svg_as_gspencil:
            bpy.ops.object.select_all(action='DESELECT')
            select_all_objects(col)
            bpy.ops.object.origin_set(type='ORIGIN_GEOMETRY', center='MEDIAN')
            group_objects(context, data, col)

        return {'FINISHED'}
